!+
! Module: evigweight_mod
!
! Authors:  R. Gastaud - SEI  CEA Saclay
!          
! Description: Weight EPIC event lists with inverse effective area
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module evigweight_mod

  use emutils_mod
  use dal
  use cal
  use param  
  use errorHandling
  use eventAttributes
  implicit none 
  
  private
  
! Names of the task parameters

    character(len=*), parameter :: inevent    = "ineventset", &
                                   newoutput  = "newoutput", &
                                   outevent   = "outeventset", &
		              weightcolname   = "weightcolname", &
                          witheffectivearea   = "witheffectivearea", &
                     withfiltertransmission   = "withfiltertransmission", &
                      withquantumefficiency   = "withquantumefficiency"

! Routine name

    character(len=*), parameter :: taskname = "evigweight"
  
! Global variables   ! Function flags
   logical :: oneffectivearea,  onfiltertransmission
   logical :: onquantumefficiency
   
  public evigweight

contains 
  
  !---------------------------------------------------------------------
  subroutine evigweight
  !---------------------------------------------------------------------
  character(len=256)  :: events_in_file, events_out_file
  logical             :: overevent , ccdnr_flag
  character(len=132)  :: weight_col_name
 
  type(DataSetT)      :: events_set
  type(TableT)        :: events_table
  type(ColumnT)       :: detx_col, dety_col, rawx_col, rawy_col, pi_col
  type(ColumnT)       :: flag_col, ccdnr_col
  type(ColumnT)       :: weight_col
  type(TableT)        :: exposu_table  ! to fill history
    
  integer(kind=int16), dimension(:), pointer   :: xdet, ydet, rawx, rawy
  integer(kind=int16), dimension(:), pointer   :: i_channels
  real(kind=single), dimension(:), pointer     :: r_channels
  real(kind=double), dimension(:), pointer     :: energies
  integer(kind=int32), dimension(:), pointer   :: quality
  
  integer(kind=int16), dimension(:), pointer      :: chipX, chipY
  integer (kind=int16)                            :: rawX0, rawY0
  real(kind=double), dimension(0:0)               :: theta, phi 
  real(kind=double), dimension(:), pointer        :: p_theta, p_phi 
  real(kind=double), dimension(0:0)               :: r_detx, r_dety, r_detz 
  real(kind=double), dimension(:), pointer        :: xmm1, ymm1, zmm1
  real(kind=double), dimension(:), pointer        :: xmm2, ymm2, zmm2
  integer(kind=int8),  dimension(:), pointer      :: ccdnr_vec
  integer(kind=int8) :: ccdnr_cur
 
  real(kind=double) :: Energy 
  real(double) :: pi, conv, step, focalLength
  real :: effarea1
  real, dimension(0:14000) :: effarea0
  integer:: nevents, iloop, coltype, energy_index
  integer(kind=int32) :: central_ccd
  real(double) :: d_central_ccd
  real(single), dimension(:), pointer :: weights
  integer(kind=int32)   :: iccd, testflag
  
  integer processed_events, status, nlowe, nhighe
  
  !**** Initialisation of the variables *****
  
  processed_events = 0
  nlowe = 0
  nhighe = 0
  oneffectivearea       = booleanParameter(witheffectivearea)
  onfiltertransmission  = booleanParameter(withfiltertransmission)
  onquantumefficiency   = booleanParameter(withquantumefficiency)
  weight_col_name = stringParameter(weightcolname)

  events_in_file = stringParameter(inevent)
  if (len_trim(events_in_file) == 0) then 
      call error("getParamvalues01","Blank input event file name")
      write(errstr,*) "No input file"
  endif
  overevent = booleanParameter(newoutput)
  if (overevent) then  ! new output
       events_out_file = stringParameter(outevent)
       call copyDataSet(events_in_file, events_out_file)
       write(errstr,*) "Input file ", trim(events_in_file), &
                                   " copied to ", trim(events_out_file)
       call message(AppMsg,Sparse)
  else   ! modify the input
       events_out_file = events_in_file
       call message(AppMsg,Sparse,"The input file will be overwritten")       
  endif
  
  testflag =   EVATT_OUT_OF_CCD_WINDOW  & ! Flag for rawx=0
             + EVATT_OUT_OF_FOV         & ! Flag out of field of view
	     + EVATT_OUTSIDE_THRESHOLDS   ! Flag energy outside of limits
 

  !**** Open the events set and Read the needed columns *****
  
  events_set = dataSet(events_out_file, Modify)
  events_table = table(events_set, 0) !  first table
  
  nevents = numberOfRows(events_table)
  write(errstr,*) "nevents =", nevents
  call message(AppMsg,Sparse)
  allocate(energies(0:nevents-1))    
  
  detx_col = column(events_table, "DETX", READ)
  xdet => int16Data(detx_col)
  dety_col = column(events_table, "DETY", READ)
  ydet => int16Data(dety_col)
  rawx_col = column(events_table, "RAWX", READ)
  rawx => int16Data(rawx_col)
  rawy_col = column(events_table, "RAWY", READ)
  rawy => int16Data(rawy_col)
  flag_col = column(events_table, "FLAG", READ)
  quality => int32Data(flag_col)
  
  pi_col = column(events_table, "PI", READ)  
  coltype = dataType(pi_col)
  select case(coltype)
         case (Integer16)
            i_channels => int16Data(pi_col)
             energies = i_channels
         case (Real32)
            r_channels => real32Data(pi_col)
            energies = r_channels
         case default
          write(errstr,*) "input column PI has an unsupported type ", &
	                  coltype, ". Contact developer"
          call error("copyData01")
   end select
  
  if (hasColumn(events_table, "CCDNR")) then
       ccdnr_col = column(events_table, "CCDNR", READ)
       ccdnr_vec => int8Data(ccdnr_col)
       ccdnr_flag = .true. 
       ccdnr_cur  = ccdnr_vec(0)
       write(errstr,*) " CCDNR exist and ccdnr_cur=", ccdnr_cur
       call message(AppMsg,Sparse)
  else
       ccdnr_flag = .false.
       call message(AppMsg,Sparse, " NO CCDNR")       
  endif
  
  if (hasColumn(events_table, weight_col_name)) then
       weight_col = column(events_table, weight_col_name, Modify)
       call message(AppMsg,Sparse, " modify weight column")         
  else
       weight_col = addColumn(events_table, weight_col_name, Real32,  &
                       comment="ratio of area center/offaxis")
  endif		       
  weights => real32Data(weight_col)
  
       
  !**** set the calibration ***** 
  ! we can not use  CAL_setState(events_in_file) for merged file
  ! write(*,*) "cal set state begin"
  if(ccdnr_flag) then
      iccd = ccdnr_cur
      status = change_ccdnr(iccd, events_set)
  else
      call CAL_setState(events_in_file)
      call CAL_getState(ccdChipId=iccd)
  endif
  
  !write(*,*) "CCD chip in cal = ",iccd
  

  !**** Compute the area at the center for each energy *****    

  call CAL_getMiscellaneousData('OPTICS_CCD', d_central_ccd)  
  central_ccd = d_central_ccd   ! central ccd is stored as a real !!!
  write(errstr,*) "central_ccd", central_ccd
  call message(AppMsg,Verbose)
  ! the central EMOS CCD is #1 (symbolic name: EMOS_CCD_CENTRAL)
  
  status = change_ccdnr(central_ccd, events_set)
 
  theta(0) = 0.
  phi(0)   = 0.
  call CAL_telCoordToCamCoord2(theta, phi, xmm2, ymm2, zmm2)  
  call CAL_camCoord2ToCamCoord1(xmm2, ymm2, zmm2,xmm1, ymm1, zmm1) 
  call CAL_camCoord1ToChipCoord(xmm1, ymm1, zmm1, chipX, chipY)
  rawX0 = chipX(0)
  rawY0 = chipY(0)
  write(errstr,*) "rawX0 =", rawX0, "rawY0 =", rawY0
  call message(AppMsg,Noisy)

  call CAL_getMiscellaneousData('FOCAL_LENGTH',focalLength)
  pi = acos(0.d0) * 2.d0
  step = 1./real64Attribute(detx_col,"TCDLT")  ! pixel size in x 
  conv = step * 180.d0 / pi / focalLength
  
  do iloop = 100, 14000
      Energy = iloop
      effarea0(iloop) = effarea(Energy, 0.d0, 0.d0 , rawX0, rawY0)
  enddo
  do iloop = 0, 100
      effarea0(iloop) = effarea0(100)
  enddo
  call message(AppMsg,Verbose)
  write(errstr,*) "mineffarea0 =", minval(effarea0), "maxeffarea0 =", &
                          maxval(effarea0)
  			  
  write(errstr,*) "step=", step, "focalLength=", focalLength, &
                          "conv=", conv
  call message(AppMsg,Verbose)
  
  !**** Compute the area for each event ***** 
  
  status = change_ccdnr(iccd, events_set)  ! back to the ccd
    
  !nevents=  2964 
  do iloop = 0, nevents-1
     Energy = energies(iloop)
     if (iand(quality(iloop),testflag) == 0) then
        processed_events = processed_events + 1 
        if(ccdnr_flag) then 
	   if(ccdnr_cur .ne. ccdnr_vec(iloop)) then
	      ccdnr_cur = ccdnr_vec(iloop)
	      iccd = ccdnr_cur  ! change the format 8 bits to 16 bits
	      status = change_ccdnr(iccd, events_set)
	      write(errstr,*) "change ccd ", iloop, ccdnr_cur
	      call message(AppMsg,Sparse)
	   endif
	endif
        r_detx(0) = xdet(iloop)/conv
        r_dety(0) = ydet(iloop)/conv
	r_detz(0) = 0.
        call CAL_camCoord2ToTelCoord(r_detx, r_dety, r_detz, p_theta, p_phi)
	theta(0) = p_theta(0)
	phi(0)   = p_phi(0)
	call CAL_releaseMemory(p_theta)
        call CAL_releaseMemory(p_phi) 
	energy_index = Energy

	if (energy_index < 100) then
!	    write(errstr,*) "energy too low", energy_index, iloop
!	    call message(AppMsg,Noisy)
            nlowe = nlowe + 1
	    energy_index = 100
	    Energy = 100.
            effarea1 = effarea(Energy,theta(0),phi(0),rawx(iloop),rawy(iloop))
	    weights(iloop) = effarea0(energy_index)/effarea1

	else if (energy_index >= 14000) then
!	    write(errstr,*) "energy too big", energy_index, iloop
!	    call message(AppMsg,Noisy)
            nhighe = nhighe + 1
	    energy_index = 14000 
	    weights(iloop) = 1.
	    effarea1 = 1.

	else
!            write(errstr,*) "energy ==", energy_index, iloop,&
!	            "theta=", theta/3600,"phi=", phi/pi*180.d0, &
!                    "r_detx =", r_detx, "r_dety=", r_dety
!	    call message(AppMsg,Noisy)
            effarea1 = effarea(Energy,theta(0),phi(0),rawx(iloop),rawy(iloop))
	    weights(iloop) = effarea0(energy_index)/effarea1
	endif

	if (effarea1 <= 0) then	   
 	   write(errstr,*) "effarea1 negative", effarea1, iloop, &
	               energy_index, "theta=", theta/3600,"phi=", phi/pi*180.d0
	   call message(AppMsg,Sparse)
           weights(iloop) = 0.
	endif
     else
     weights(iloop) = 0.	
     endif

	
  enddo
 
  !**** fill history ***** 
  if (oneffectivearea) &
         call keywordDone(block(events_table), taskname, witheffectivearea) 
  if (onfiltertransmission) &
         call keywordDone(block(events_table),taskname,withfiltertransmission)
  if (onquantumefficiency) &
         call keywordDone(block(events_table),taskname,withquantumefficiency )
  if (hasBlock(events_set, "EXPOSURE")) then
      call message(AppMsg,Verbose,"found  EXPOSURE ")
      exposu_table = table(events_set, "EXPOSURE")
      call addAttributes(attributable(exposu_table),attributable(events_table))
  endif 
  
  ! house keeping
  call release(events_table)
  call release( events_set )
  
  deallocate(energies)
  
  write(errstr, *) " total number of processed events ", processed_events 
  call message(AppMsg,Sparse)
  write(errstr,*) nlowe, " events with energy lower than 100 eV"
  call message(AppMsg,Verbose)
  write(errstr,*) nhighe, " events with energy higher than 14 keV"
  call message(AppMsg,Verbose)

  end subroutine evigweight

 !---------------------------------------------------------------------
  real function effarea(Energy, theta, phi, rawx, rawy)
  !  compute the effective area for a given energy and position
  real(kind=double)  , intent(in) :: Energy  
  real(kind=double)   ,intent(in) :: theta, phi
  integer (kind=int16) , intent(in) :: rawx, rawy
  
  real(kind=double)            :: dtheta, dphi  
  
  effarea = 1.
  dtheta = theta
  dphi   = phi
 
   
  if (oneffectivearea) then
        effarea = effarea*CAL_getEffectiveArea(Energy, dtheta, dphi)
  endif
 
  if (onfiltertransmission) then
        effarea = effarea*CAL_getFilterTransmission(Energy, dtheta, dphi)
  endif
	
  if (onquantumefficiency) then
        effarea = effarea*CAL_getQuantumEfficiency(Energy, rawx, rawy)
	! pattern by default : 0
  endif
	
  
  end function effarea
 !---------------------------------------------------------------------
 
 !---------------------------------------------------------------------
  integer function change_ccdnr(iccd, events_set)
  !  set the new ccdrn in the calibration set
  !  add the history in the extension exposu
  
  integer        , intent(in) :: iccd
  type(DataSetT) , intent(in) :: events_set 
  
  type(BlockT)     :: exposu_block
  character(len=8) :: exposu_tab_name

    
  change_ccdnr = 0
  write(exposu_tab_name, '(A, I2.2)') 'EXPOSU', iccd
  if (hasBlock(events_set, exposu_tab_name)) then
      !write(*,*) "found ", exposu_tab_name 
      exposu_block = block(events_set, exposu_tab_name)
      call CAL_setState(exposu_block)
      change_ccdnr = 1
  else
      !write(*,*) "NOT found ", exposu_tab_name 	  
      call CAL_setState(ccdChipId=iccd)	
      change_ccdnr = 2		       
  endif
  
  end function change_ccdnr
 
 !---------------------------------------------------------------------
 
 
end module evigweight_mod
